@@ -1,9 +1,9 @@
 # Name:: FormatR
 # Description:: Perl like formats for ruby
 # Author:: Paul Rubel (rubel@crhc.uiuc.edu)
-# Release:: 1.05
+# Release:: 1.06
 # Homepage:: http://www.crhc.uiuc.edu/~rubel/FormatR.html
-# Date:: 17 April 2002
+# Date:: 29 July 2002
 # License:: You can redistribute it and/or modify it under the same term as Ruby.
 #           Copyright (c) 2002  Paul Rubel
 #
@@ -11,6 +11,7 @@
 #     IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 #     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 #     PURPOSE.
+#
 # = To Test this code:
 # Try test_format.rb with no arguments. If nothing is amiss you should
 # see OK (??/?? tests ?? asserts). This tests the format output
@@ -92,7 +93,7 @@
 #
 # * A line that starts with ~ will be suppressed if it will be blank
 #
-# * A line that starts with ~~ will repeat until it is blank, be sure
+# * A line that contains ~~ will repeat until it is blank, be sure
 #   to use this feature with at least one field starting with a ^.
 #
 # === Scientific formats of the form @.#G##, @.#g##, @.#E##, and @.#e##
@@ -117,8 +118,91 @@
 # * More examples of using the scientific formats can be found in test_format.rb
 #
 #
+# = Reading in output printed by formats, FormatR::FormatReader
+#
+# The class FormatR::FormatReader can be used to read in text that has
+# been output with a given format and attepmts to extract the values of the
+# variables used as the input. It does a good job of simple formats, I'm sure
+# that there are complex ones that can confuse it. Multi-line formats are supported
+# but as the program can't be sure what the initial input looked like, and how it 
+# was broken across lines, every piece of a line is made to have at least one
+# space after it.
+#
+# For example: if you had the following format:
+#
+#  ~~^<<
+#  var
+# and you fed it the string abcdef you would get the following:
+#    abc
+#    def
+# But when var was assigned to it would be var = 'abc def'
+# 
+# I don't know how to decide which is better. Perhaps an argument would help
+# 
+#
+# == The classes of variables
+#
+# It's not always possible to infer the class of the variable that made the
+# format. By not taking in a binding to compare with many variables will end
+# up as strings. Numeric formats should come out as numbers but all others will
+# be strings and will need to be converted manually.
+#
+# == Using FormatR::FormatReader 
+#
+# Using the FormatReader is relatively simple. You pass in a format to the
+# constructor and then call readFormat and give in an array of formatted text.
+# It will return a hash with the key/value pairs of the variables in the
+# format. 
+#
+# For example:
+#
+# 
+#    f = []
+#    # make a format
+#    f.push( '<?xml version="1.0"?>' )
+#    f.push( '@@@ Blah @@@ }Blah @< @|| @#.#' )
+#    f.push( 'var_one,var_one,var_one,var_one,var_one,var_one,' +
+#           ' var_two, var_three, var_four')
+#    f.push( '@<<< @<<<')
+#    f.push( 'var_one,var_one')
+#    format = Format.new(f)
+# 
+#    #set values and print it out.
+#    var_one, var_two, var_three, var_four = 1, 2, 3, 4.3
+#    output_filename = "format_testfile12"
+#    File.open( output_filename, File::CREAT | File::WRONLY | File::TRUNC ) { |file|
+#      format.io = file
+#      format.printFormat(binding) 
+#    }
+#    # read in the output
+#    output = []
+#    File.open( output_filename ){ |file|
+#      output = file.readlines()
+#    }
+#   
+#   # make a new FormatReader
+#   reader = FormatReader.new (format)
+#   # read in the values
+#    res = reader.readFormat (output)
+# 
+#   # Check that the values are correct
+#    assert (res['var_one'] == var_one.to_s)
+#    assert (res['var_two'] == var_two.to_s)
+#    assert (res['var_three'] == var_three.to_s)
+#    assert (res['var_four'] == var_four)
+#
+#
+# 
+#
 # = Changes:
 # 
+# ==1.06
+# * I thought that the ~ had to be in the front of the picture line,
+#   this isn't so. If you place the ~~ anywhere in the line it will
+#   repeat until the line is empty.
+#
+# * Added the FormatReader to read in formatted text and get values back
+#
 # ==1.05
 # * Hugh Sasse sent in a patch to clean up warnings. I was sloppy with my  
 #   spacing but hopefully have learned better. Thanks Hugh!
@@ -168,9 +252,10 @@
 # 
 # * Move test_format.rb over to runit. 
 #
-# * Added functionality so that if you pass in a format string, or array of strings
-#   to setTop or setBottom it does the right thing. This way you don't need to make the 
-#   extra formats just to pass them in. 
+# * Added functionality so that if you pass in a format string, or
+#   array of strings to setTop or setBottom it does the right
+#   thing. This way you don't need to make the extra formats just to
+#   pass them in.
 #
 #
 # ==1.02
@@ -288,12 +373,12 @@ module FormatR
   # This class holds a single block of text, either something
   # unchanging or a picture element of some format.
   class FormatEntry
-    attr_accessor :val
+    attr_accessor :val, :unchanging
 
-    def initialize (val, space)
-      @space = space
+    def initialize (val, unchanging)
+      @unchanging = unchanging
       @val = val
-      if (!space)
+      if (!unchanging)
         s = val.size - 1
         
         if (val =~ /[@^][<]{#{s},#{s}}/)
@@ -313,8 +398,8 @@ module FormatR
     end
     
     # is this just unchanging characters
-    def isUnchanging ()
-      return @space
+    def isUnchanging? ()
+      return @unchanging
     end
 
     # give back the string passed through the appropriate formatter
@@ -322,7 +407,15 @@ module FormatR
       result = @formatter.formatString(string, var_name, aBinding)
       return result
     end
-  end
+
+    # show our values
+    def to_s ()
+      output =  "'" + @val + "'   unchanging:" + @unchanging.to_s
+      #(output << " formatter:" + @formatter.class.to_s) if (!@unchanging)
+      output
+    end
+
+  end # end of class FormatEntry
 
   # This is the base class for all the formats, <,>,|, and # of the
   # @ or ^ persuasion. It keeps track of filled variables and the length
@@ -362,7 +455,6 @@ module FormatR
       return changeVarValue(var_value, var_name, binding)
     end
 
-
   end
 
   # this format doesn't care if it's a @ or an ^, it acts the same and doesn't chop things
@@ -582,7 +674,8 @@ module FormatR
     # need to pass an empty format in as the first argument or a
     # nil. The output defaults to standard out but can be changed with
     # the Format.io= method. 
-    #  The format is a string in the style of a perl format or an array
+    #
+    # The format is a string in the style of a perl format or an array
     # of strings each of which is a line of a perl format. The passed
     # in format contains multiple lines, picture lines and argument
     # lines. A picture line can contain any text but if it contains an
@@ -653,6 +746,36 @@ module FormatR
       setBottom(bottom_format) if bottom_format
     end
 
+    # print out all the values we're holding for pictures
+    def showPictureLine ()
+      @picture_lines.each do |line|
+        puts "line:"
+        line.each do |element|
+          puts "  #{element.to_s} "
+        end
+      end
+    end
+
+    # return an Array of picture line FormatHolder s
+    def getPictureLines ()
+      output = Array.new
+      @picture_lines.each_index do |i|
+        line = @picture_lines[i]
+        vars = @vars[i].dup
+        output_line = FormatHolder.new
+        output_line.repeat = line.repeat
+        line.each do |element|
+          val = element.val
+          var_name = nil
+          var_name = vars.shift() unless (element.unchanging)
+          var_name.strip! unless (element.unchanging)
+          #puts "got val #{val} var #{var_name}."
+          output_line.push( [val,var_name] ) unless val == ""
+        end
+        output.push( output_line )
+      end
+      output
+    end
 
     # Things you shouldn't have to deal with.
     private
@@ -680,10 +803,9 @@ module FormatR
       nonFormats = getNonFormats(line)
       formats = getFormats(line)
       a = FormatHolder.new()
-      a.repeat = (line =~ /^~~.*/) ? true :false
-      a.suppress = (line =~ /^~.*/) ? true : false
+      a.repeat = (line =~ /.*~~.*/) ? true :false
+      a.suppress = (line =~ /.*~.*/) ? true : false
       nonFormats.each_index {|i|
-        puts "adding non #{nonFormats[i]} " if (nonFormats[i] && nonFormats[i] =~ /.*oo/)
         a.push( FormatEntry.new( nonFormats[i], true )) if ( nonFormats[i] )
         a.push( FormatEntry.new( formats[i], false )) if ( formats[i] )
       }
@@ -882,8 +1004,9 @@ module FormatR
       return false
     end
 
+
     # The workhorse of printFormat, buffer the format contained and
-    #  then try to print it out
+    # then try to print it out
     def printBodyFormat (io)
       @picture_lines.each_index {|i| # There isn't a 1-1 correspondence here anymore!
         line = @picture_lines[i]
@@ -893,7 +1016,7 @@ module FormatR
           outputLine = ""
           vars = @vars[i].dup if @vars[i]
           line.each {|item|
-            if (item.isUnchanging()) 
+            if (item.isUnchanging?()) 
               outputLine << "#{item.val}"
             else
               begin
@@ -929,4 +1052,130 @@ module FormatR
     attr_accessor :suppress, :repeat
   end
 
+
+# This class takes in a format and instead of writing out the values
+# variables under the given format will read in formatted text and give
+# the values of variables as specified in the given format. 
+  class FormatReader
+    
+    # Make a FormatReader given a format
+    def initialize (format) 
+      @pictures = format.getPictureLines()
+      @var_values = Hash.new
+    end
+    
+    # Given the output from a format return a hash with the values
+    # of the variables given in the input mapped to the variables in 
+    # the format. 
+    def readFormat (output) 
+      @var_values = Hash.new
+      output_line = 0
+      @pictures.each_index {|i|
+        repeat = true
+        while (repeat)
+          found_match = setLine( @pictures[i], output[output_line] )
+          repeat = false #default to stopping
+          if (found_match)
+            output_line += 1
+          end
+          #we may need to repeat if it's a ~~ line
+          if (@pictures[i].repeat() && found_match)
+              repeat = true
+          end
+        end #while
+      }
+      @var_values
+    end
+    
+    private 
+    
+    #given a picture line remove spaces that have probably been added
+    #as padding.
+    def removeSpaces (picture, data)
+      result = data
+      if (picture.include?('<'))
+        result.gsub!( / *$/, '')
+      elsif (picture.include?('>'))
+        result.gsub!( /^ */, '' )
+      elsif (picture.include?('|'))
+        result = result.strip
+      end
+      result
+    end
+
+    # Put a value into @var_values. Guess at the type
+    def saveVar (picture, name, data)
+      to_save = removeSpaces( picture, data )
+      #numbers
+      if (picture.include?('#'))
+        if (picture.include?('.'))
+          @var_values[name] = to_save.to_f
+        else
+          @var_values[name] = to_save.to_i
+        end
+      # continuation lines
+      elsif (picture =~ /^\^.*/) 
+        curval = @var_values[name]
+        curval ||= ""
+        # we assume you want a space between the two things split
+        if (curval != "" && to_save =~ /^[^\s]*/)
+          to_save =  " " + to_save 
+        end
+        #puts "setting #{name} to #{curval + to_save}"
+        @var_values[name] = curval + to_save
+      else
+        @var_values[name] = to_save
+      end
+    end
+
+    # Given a regexp some variable names and output fill the
+    # variables in with data from the output according the the format
+    # in the regexp. Will return false if no matches can
+    # be found, true if matches were found.
+    def findVars (regexp_string, vars, output) 
+      r = Regexp.new( regexp_string )
+      match_data = r.match( output )
+      output = Hash.new
+      if (!match_data.nil?)
+        matches = match_data.to_a
+        matches = matches[1,matches.length]
+        matches.each_with_index do |data, index|
+          str, var = vars[index]
+          #puts "match str: #{str}   #{var} = #{data}"
+          saveVar( str, var, data )
+        end
+      else
+        #no match data
+        return false
+      end
+      true
+    end
+
+    # place the values contained in the output specified by the
+    # picture line in @var_values
+    def setLine (picture_line, output)
+      regexp_string = ""
+      vars = []
+      picture_line.each_with_index {|elem, index|
+        str, var = elem
+        if (var.nil?) 
+          regexp_string << Regexp.escape(str)
+          
+        else # capture the variable's section of the string
+          vars.push( elem )
+          if (index != picture_line.length - 1)
+            regexp_string << '(' 
+            regexp_string << '.' * str.length
+            regexp_string << ')' 
+          else
+            regexp_string << '(.*)'
+          end
+        end
+      }
+      
+      findVars( regexp_string, vars, output )
+    end
+
+  end
+  
 end
